import sys
import numpy as np


def calculate_asymptotic_training_length(forget: float):
    if forget == 0:
        # Return the highest possible value that is no infinity
        return sys.maxsize
    else:
        return 1 / forget


def calculate_effective_training_length(forget: float, n_obs: int):
    if forget == 0:
        return n_obs
    else:
        return (1 - (1 - forget) ** n_obs) / forget


def online_mean_update(avg: float, value: float, forget: float, n_seen: int):

    n_asymmptotic = calculate_asymptotic_training_length(forget)
    n_eff = calculate_effective_training_length(forget, n_seen)

    forget_scaled = forget * np.maximum(n_asymmptotic / n_eff, 1.0)

    diff = value - avg
    incr = forget_scaled * diff

    if forget_scaled > 0:
        new_avg = avg + incr
    else:
        new_avg = avg + diff / n_seen

    return new_avg



import numpy as np
import scipy.stats as st

from rolch.abc import Distribution
from rolch.link import IdentityLink, LogLink


class DistributionJSU(Distribution):
    """
    Corresponds to GAMLSS JSUo() and scipy.stats.johnsonsu()

    Distribution parameters:
    0 : Location
    1 : Scale (close to standard deviation)
    2 : Skewness
    3 : Tail behaviour
    """

    def __init__(
        self,
        loc_link=IdentityLink(),
        scale_link=LogLink(),
        shape_link=IdentityLink(),
        tail_link=LogLink(),
    ):
        self.n_params = 4
        self.loc_link = loc_link
        self.scale_link = scale_link
        self.shape_link = shape_link
        self.tail_link = tail_link
        self.links = [
            self.loc_link,
            self.scale_link,
            self.shape_link,  # skew
            self.tail_link,  # tail
        ]

    def theta_to_params(self, theta):
        mu = theta[:, 0]
        sigma = theta[:, 1]
        nu = theta[:, 2]
        tau = theta[:, 3]
        return mu, sigma, nu, tau

    def dl1_dp1(self, y, theta, param=0):
        mu, sigma, nu, tau = self.theta_to_params(theta)

        if param == 0:
            # MU
            z = (y - mu) / sigma
            r = nu + tau * np.arcsinh(z)
            dldm = (z / (sigma * (z * z + 1))) + (
                (r * tau) / (sigma * np.sqrt(z * z + 1))
            )
            return dldm

        if param == 1:
            # SIGMA
            z = (y - mu) / sigma
            r = nu + tau * np.arcsinh(z)
            dldd = (-1 / (sigma * (z * z + 1))) + (
                (r * tau * z) / (sigma * np.sqrt(z * z + 1))
            )
            return dldd

        if param == 2:
            # nu
            z = (y - mu) / sigma
            r = nu + tau * np.arcsinh(z)
            dldv = -r
            return dldv

        if param == 3:
            # tau
            z = (y - mu) / sigma
            r = nu + tau * np.arcsinh(z)
            dldt = (1 + r * nu - r * r) / tau
            return dldt

    def dl2_dp2(self, y, theta, param=0):
        mu, sigma, nu, tau = self.theta_to_params(theta)
        if param == 0:
            # MU
            z = (y - mu) / sigma
            r = nu + tau * np.arcsinh(z)
            dldm = (z / (sigma * (z * z + 1))) + (
                (r * tau) / (sigma * np.sqrt(z * z + 1))
            )
            d2ldm2 = -dldm * dldm
            # d2ldm2 = np.max(1e-15, d2ldm2)
            return d2ldm2

        if param == 1:
            # SIGMA
            z = (y - mu) / sigma
            r = nu + tau * np.arcsinh(z)
            dldd = (-1 / (sigma * (z * z + 1))) + (
                (r * tau * z) / (sigma * np.sqrt(z * z + 1))
            )
            d2ldd2 = -(dldd * dldd)
            # d2ldd2 = np.max(d2ldd2, -1e-15)
            return d2ldd2

        if param == 2:
            # TAIL
            z = (y - mu) / sigma
            r = nu + tau * np.arcsinh(z)
            d2ldv2 = -(r * r)
            # d2ldv2 = np.max(d2ldv2 < -1e-15)
            return d2ldv2
        if param == 3:
            z = (y - mu) / sigma
            r = nu + tau * np.arcsinh(z)
            dldt = (1 + r * nu - r * r) / tau
            d2ldt2 = -dldt * dldt
            # d2ldt2 = np.max(d2ldt2, -1e-15)
            return d2ldt2

    def dl2_dpp(self, y, theta, params=(0, 1)):
        mu, sigma, nu, tau = self.theta_to_params(theta)
        if sorted(params) == [0, 1]:
            z = (y - mu) / sigma
            r = nu + tau * np.arcsinh(z)
            dldm = (z / (sigma * (z * z + 1))) + (
                (r * tau) / (sigma * np.sqrt(z * z + 1))
            )
            dldd = (-1 / (sigma * (z * z + 1))) + (
                (r * tau * z) / (sigma * np.sqrt(z * z + 1))
            )
            d2ldmdd = -(dldm * dldd)
            return d2ldmdd

        if sorted(params) == [0, 2]:
            z = (y - mu) / sigma
            r = nu + tau * np.arcsinh(z)
            dldm = (z / (sigma * (z * z + 1))) + (
                (r * tau) / (sigma * np.sqrt(z * z + 1))
            )
            dldv = -r
            d2ldmdv = -(dldm * dldv)
            return d2ldmdv

        if sorted(params) == [0, 3]:
            z = (y - mu) / sigma
            r = nu + tau * np.arcsinh(z)
            dldm = (z / (sigma * (z * z + 1))) + (
                (r * tau) / (sigma * np.sqrt(z * z + 1))
            )
            dldt = (1 + r * nu - r * r) / tau
            d2ldmdt = -(dldm * dldt)
            return d2ldmdt

        if sorted(params) == [1, 2]:
            z = (y - mu) / sigma
            r = nu + tau * np.arcsinh(z)
            dldd = (-1 / (sigma * (z * z + 1))) + (
                (r * tau * z) / (sigma * np.sqrt(z * z + 1))
            )
            dldv = -r
            d2ldddv = -(dldd * dldv)
            return d2ldddv

        if sorted(params) == [1, 3]:
            z = (y - mu) / sigma
            r = nu + tau * np.arcsinh(z)
            dldd = (-1 / (sigma * (z * z + 1))) + (
                (r * tau * z) / (sigma * np.sqrt(z * z + 1))
            )
            dldt = (1 + r * nu - r * r) / tau
            d2ldddt = -(dldd * dldt)
            return d2ldddt

        if sorted(params) == [2, 3]:
            z = (y - mu) / sigma
            r = nu + tau * np.arcsinh(z)
            dldv = -r
            dldt = (1 + r * nu - r * r) / tau
            d2ldvdt = -(dldv * dldt)
            return d2ldvdt

    def link_function(self, y, param=0):
        return self.links[param].link(y)

    def link_inverse(self, y, param=0):
        return self.links[param].inverse(y)

    def link_derivative(self, y, param=0):
        return self.links[param].derivative(y)

    def initial_values(self, y, param=0, axis=None):
        if param == 0:
            return (y + np.mean(y, axis=None)) / 2
        if param == 1:
            return np.repeat(np.std(y, axis=None), y.shape[0])
        if param == 2:
            return np.full_like(y, 0)
        if param == 3:
            return np.full_like(y, 10)

    def cdf(self, y, theta):
        mu, sigma, nu, tau = self.theta_to_params(theta)
        return st.johnsonsu(
            loc=mu,
            scale=sigma,
            a=nu,
            b=tau,
        ).cdf(y)

    def pdf(self, y, theta):
        mu, sigma, nu, tau = self.theta_to_params(theta)
        return st.johnsonsu(
            loc=mu,
            scale=sigma,
            a=nu,
            b=tau,
        ).pdf(y)

    def ppf(self, q, theta):
        mu, sigma, nu, tau = self.theta_to_params(theta)
        return st.johnsonsu(
            loc=mu,
            scale=sigma,
            a=nu,
            b=tau,
        ).ppf(q)

    def rvs(self, size, theta):
        mu, sigma, nu, tau = self.theta_to_params(theta)
        return (
            st.johnsonsu(
                loc=mu,
                scale=sigma,
                a=nu,
                b=tau,
            )
            .rvs((size, theta.shape[0]))
            .T
        )



from abc import ABC, abstractmethod
from typing import Tuple, Union

import numpy as np


class LinkFunction(ABC):
    """The base class for the link functions."""

    @abstractmethod
    def link(self, x: np.ndarray) -> np.ndarray:
        """Calculate the Link"""

    @abstractmethod
    def inverse(self, x: np.ndarray) -> np.ndarray:
        """Calculate the inverse of the link function"""

    @abstractmethod
    def derivative(self, x: np.ndarray) -> np.ndarray:
        """Calculate the first derivative of the link function"""


class Distribution(ABC):

    @abstractmethod
    def theta_to_params(self, theta: np.ndarray) -> Tuple:
        """Take the fitted values and return tuple of vectors for distribution parameters."""

    @abstractmethod
    def dl1_dp1(self, y: np.ndarray, theta: np.ndarray, param: int) -> np.ndarray:
        """Take the first derivative of the likelihood function with respect to the param."""

    @abstractmethod
    def dl2_dp2(self, y: np.ndarray, theta: np.ndarray, param: int) -> np.ndarray:
        """Take the second derivative of the likelihood function with respect to the param."""

    @abstractmethod
    def dl2_dpp(
        self, y: np.ndarray, theta: np.ndarray, params: Tuple[int, int]
    ) -> np.ndarray:
        """Take the first derivative of the likelihood function with respect to both parameters."""

    @abstractmethod
    def link_function(self, y: np.ndarray, param: int = 0) -> np.ndarray:
        """Apply the link function for param on y."""

    @abstractmethod
    def link_inverse(self, y: np.ndarray, param: int = 0) -> np.ndarray:
        """Apply the inverse of the link function for param on y."""

    @abstractmethod
    def link_derivative(self, y: np.ndarray, param: int = 0) -> np.ndarray:
        """Apply the derivative of the link function for param on y."""

    @abstractmethod
    def initial_values(
        self, y: np.ndarray, param: int = 0, axis: int = None
    ) -> np.ndarray:
        """Calculate the initial values for the GAMLSS fit."""

    @abstractmethod
    def cdf(self, y: np.ndarray, theta: np.ndarray) -> np.ndarray:
        """The cumulative density function."""

    @abstractmethod
    def pdf(self, y: np.ndarray, theta: np.ndarray) -> np.ndarray:
        """The density function."""

    @abstractmethod
    def ppf(self, q: np.ndarray, theta: np.ndarray) -> np.ndarray:
        """The percentage point or quantile function of the distribution."""

    @abstractmethod
    def rvs(self, size: Union[int, Tuple], theta: np.ndarray) -> np.ndarray:
        """Draw random samples of shape size."""



from typing import Dict, Literal, Optional

import numpy as np

from rolch.coordinate_descent import (
    DEFAULT_ESTIMATOR_KWARGS,
    online_coordinate_descent_path,
)
from rolch.gram import init_gram, init_y_gram, update_gram, update_y_gram
from rolch.information_criteria import select_best_model_by_information_criterion
from rolch.scaler import OnlineScaler
from rolch.utils import calculate_effective_training_length


class OnlineLasso:

    "Simple Online Lasso regression for the expected value."

    def __init__(
        self,
        forget=0,
        ic: Literal["aic", "bic", "hqc", "max"] = "bic",
        scale_inputs: bool = True,
        intercept_in_design: bool = True,
        lambda_n: int = 100,
        lambda_eps: float = 1e-4,
        estimation_kwargs: Optional[Dict] = None,
    ):
        self.ic = ic
        self.forget = forget
        self.lambda_n = lambda_n
        self.lambda_eps = lambda_eps
        self.intercept_in_design = intercept_in_design
        self.scaler = OnlineScaler(
            forget=forget, do_scale=scale_inputs, intercept=intercept_in_design
        )
        self.intercept = True

        for i, attribute in DEFAULT_ESTIMATOR_KWARGS.items():
            if (estimation_kwargs is not None) and (i in estimation_kwargs.keys()):
                setattr(self, i, estimation_kwargs[i])
            elif i == "lambda_eps":
                setattr(self, i, estimation_kwargs[i][0])
            else:
                setattr(self, i, attribute)

    def fit(self, y, X, sample_weights=None, beta_bounds=None):
        """Fit the regression model."""
        self.N = X.shape[0]
        self.J = X.shape[1]

        if sample_weights is None:
            sample_weights = np.ones_like(y)

        if beta_bounds is None:
            self.beta_bounds = (np.repeat(-np.inf, self.J), np.repeat(np.inf, self.J))
        else:
            self.beta_bounds = beta_bounds

        self.training_length = calculate_effective_training_length(self.forget, self.N)

        self.scaler.fit(X)
        X_scaled = self.scaler.transform(X)

        self.beta_path = np.zeros((self.lambda_n,))
        self.x_gram = init_gram(X_scaled, sample_weights, self.forget)
        self.y_gram = init_y_gram(X_scaled, y, sample_weights, self.forget)

        self.beta_path = np.zeros((self.lambda_n, self.J))
        self.is_regularized = np.repeat(True, self.J)
        self.is_regularized[0] = self.intercept_in_design == False

        intercept = (
            self.y_gram[~self.is_regularized]
            / np.diag(self.x_gram)[~self.is_regularized]
        )

        self.lambda_max = np.max(
            np.abs(self.y_gram.flatten() - self.x_gram[0] * intercept)
        )
        self.lambda_path = np.geomspace(
            self.lambda_max, self.lambda_max * self.lambda_eps, self.lambda_n
        )
        self.beta_path = online_coordinate_descent_path(
            x_gram=self.x_gram,
            y_gram=self.y_gram.flatten(),
            beta_path=self.beta_path,
            lambda_path=self.lambda_path,
            is_regularized=self.is_regularized,
            beta_lower_bound=self.beta_bounds[0],
            beta_upper_bound=self.beta_bounds[1],
            which_start_value="previous_lambda",
            selection=self.selection,
            tolerance=self.tolerance,
            max_iterations=self.max_iterations,
        )[0]
        residuals = np.expand_dims(y, -1) - X_scaled @ self.beta_path.T
        self.rss = np.sum(residuals**2, axis=0)
        n_params = np.sum(~np.isclose(self.beta_path, 0), axis=1)
        best_ic = select_best_model_by_information_criterion(
            self.training_length, n_params, self.rss, self.ic
        )
        self.beta = self.beta_path[best_ic, :]

    def update(self, y, X, sample_weights=None):
        """Update the regression model"""
        self.N += X.shape[0]
        self.training_length = calculate_effective_training_length(self.forget, self.N)
        self.scaler.partial_fit(X)
        X_scaled = self.scaler.transform(X)

        self.x_gram = update_gram(self.x_gram, X_scaled, self.forget, sample_weights)
        self.y_gram = update_y_gram(
            self.y_gram, X_scaled, y, self.forget, sample_weights
        )

        intercept = (
            self.y_gram[~self.is_regularized]
            / np.diag(self.x_gram)[~self.is_regularized]
        )

        self.lambda_max = np.max(
            np.abs(self.y_gram.flatten() - self.x_gram[0] * intercept)
        )
        self.lambda_path = np.geomspace(
            self.lambda_max, self.lambda_max * self.lambda_eps, self.lambda_n
        )
        self.beta_path = online_coordinate_descent_path(
            x_gram=self.x_gram,
            y_gram=self.y_gram.flatten(),
            beta_path=self.beta_path,
            lambda_path=self.lambda_path,
            is_regularized=self.is_regularized,
            beta_lower_bound=self.beta_bounds[0],
            beta_upper_bound=self.beta_bounds[1],
            which_start_value=self.start_value,
            selection=self.selection,
            tolerance=self.tolerance,
            max_iterations=self.max_iterations,
        )[0]

        residuals = np.expand_dims(y, -1) - X_scaled @ self.beta_path.T
        self.rss = (1 - self.forget) * self.rss + np.sum(residuals**2, axis=0)
        n_params = np.sum(~np.isclose(self.beta_path, 0), axis=1)
        best_ic = select_best_model_by_information_criterion(
            self.training_length, n_params, self.rss, self.ic
        )
        self.beta = self.beta_path[best_ic, :]

    def predict(self, X):
        """Predict using the optimal IC selection."""
        X_scaled = self.scaler.transform(X)
        prediction = X_scaled @ self.beta.T
        return prediction

    def predict_path(self, X):
        """Predict the full LASSO path."""
        X_scaled = self.scaler.transform(X)
        prediction = X_scaled @ self.beta_path.T
        return prediction
